#ifndef DATATYPE_H
#define DATATYPE_H

#ifdef __cplusplus
extern "C" {
#endif

#include <assert.h>
#include <stdbool.h>

#include "Check.h"
#include "Types.h"

typedef enum {
    VT_INVALID = 0,
    VT_NOT_SET,
    VT_INT,
    VT_FLOAT,
    VT_POINTER,
    VT_ARRAY,
    VT_LAST
} ValueType;

#define VALUE_TYPE_BIT_SIZE 4
static_assert(VT_LAST < (1 << VALUE_TYPE_BIT_SIZE));

typedef struct {
    uint32 typeAndOffset;
    union {
        int32 intValue;
        float32 floatValue;
    } data;
} Value;

typedef enum {
    RDT_INVALID = 0,
    RDT_INT,
    RDT_FLOAT,
    RDT_VOID,
    RDT_STRUCT,
    RDT_LAST
} RawDataType;

typedef struct {
    uint32 data;
} DataType;

#define DT_TYPE_BIT_SIZE 4
#define DT_IS_POINTER_BIT_SIZE 1
#define DT_IS_ARRAY_BIT_SIZE 1
#define DT_STRUCT_ID_BIT_SIZE                                                  \
    (32 - DT_TYPE_BIT_SIZE - DT_IS_POINTER_BIT_SIZE - DT_IS_ARRAY_BIT_SIZE)
static_assert(RDT_LAST < (1 << DT_TYPE_BIT_SIZE));

#define DT_MAX_STRUCT_ID ((1 << DT_STRUCT_ID_BIT_SIZE) - 1)

#define DT_TYPE_OFFSET 0
#define DT_IS_POINTER_OFFSET (DT_TYPE_OFFSET + DT_TYPE_BIT_SIZE)
#define DT_IS_ARRAY_OFFSET (DT_IS_POINTER_OFFSET + DT_IS_POINTER_BIT_SIZE)
#define DT_STRUCT_ID_OFFSET (DT_IS_ARRAY_OFFSET + DT_IS_ARRAY_BIT_SIZE)

typedef struct {
    const char* name;
    DataType type;
    char padding[sizeof(const char*) - sizeof(DataType)];
} StructVariable;

typedef struct {
    const char* name;
    uint16 id;
    char padding[2];
    int amount;
    StructVariable* vars;
} Struct;

typedef struct {
    int capacity;
    int entries;
    Struct* data;
} Structs;

const char* vtGetName(ValueType vt);
ValueType vGetType(Value v);
check_return bool vSetType(Value* v, ValueType vt);
int32 vGetOffset(Value v);
check_return bool vSetOffset(Value* v, int32 offset);

int dtGetSize(DataType dt, const Structs* sts);

DataType dtInt(void);
DataType dtFloat(void);
DataType dtVoid(void);
DataType dtStruct(const Struct* st);
DataType dtText(void);

DataType dtToArray(DataType dt);
DataType dtRemoveArray(DataType dt);
Struct* dtGetStruct(const Structs* sts, DataType dt);

DataType dtToPointer(DataType dt);
bool dtRemovePointer(DataType* dt);

RawDataType dtGetType(DataType dt);
bool dtCompare(DataType a, DataType b);
bool dtIsInt(DataType dt);
bool dtIsFloat(DataType dt);
bool dtIsVoid(DataType dt);
bool dtIsArray(DataType dt);
bool dtIsStruct(DataType dt);
bool dtIsPointer(DataType dt);

const char* dtGetName(const Structs* sts, DataType dt);

void stAddVariable(Struct* st, const char* name, DataType type);

void stsInit(Structs* sts);
void stsDelete(Structs* sts);
Struct* stsSearch(Structs* sts, const char* name);
Struct* stsAdd(Structs* sts, const char* name);

void gstsInit(void);
void gstsDelete(void);
Structs* gstsGet(void);
Struct* gstsAdd(const char* name);

#ifdef __cplusplus
}
#endif

#endif